export { minLength };
export { maxLength };
export { min };
export { max };
export { range };

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

import * as util from './util';
import ShortId from 'shortid';

function isFilled(value) {
    return !util.isEmpty(value);
}

export function custom(validateFn, options) {
    return _extends({
        id: ShortId.generate(), // makes this rule serializeable
        isAsync: false, // validate is expected to return a Promise. Validation rule will not run if other synchronous validation rules are failing
        message: "This field is invalid.", // {string|Function} message to display if input is not valid
        validate: validateFn,
        precondition: function precondition() {
            return true;
        }, // {Function} if precondition fails, validation rule is not ran (not cached). Probably not needed if sync functions run first.
        dependsOn: [],
        isOptional: true, // don't run validation rule if _.isEqual(value,defaultValue)
        type: 'error', // "error" or "warning"
        compare: util.arrayCompare }, options);
}

export function async(validateFn, options) {
    return custom(validateFn, _extends({ isAsync: true }, options));
}

export var required = custom(isFilled, {
    id: 'required',
    message: 'This field is required.',
    isOptional: false
});

function minLength(length) {
    var _message = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : function (value, length) {
        return 'Please enter at least ' + length + ' characters (' + (length - value.length) + ' more).';
    };

    return custom(function (val) {
        return val.length >= length;
    }, { id: 'minLength', message: function message(value) {
            return _message(value, length);
        } });
}

function maxLength(length) {
    var _message2 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : function (value, length) {
        return 'Please enter at most ' + length + ' characters. You\'ve entered ' + value.length + '.';
    };

    return custom(function (val) {
        return val.length <= length;
    }, { id: 'maxLength', message: function message(value) {
            return _message2(value, length);
        } });
}

export var email = custom(function (value) {
    return (/^(([^<>()\[\].,;:\s@"]+(\.[^<>()\[\].,;:\s@"]+)*)|(".+"))@(([^<>()[\].,;:\s@"]+\.)+[^<>()[\].,;:\s@"]{2,})$/i.test(value)
    );
}, {
    id: 'email',
    message: 'Please enter a valid email address.'
});

function min(min) {
    var _message3 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : function (value, min) {
        return 'Please enter a value greater than or equal to ' + min + '.';
    };

    return custom(function (value) {
        return !isNaN(value) && value >= min;
    }, { id: 'min', message: function message(value) {
            return _message3(value, min);
        } });
}

function max(max) {
    var _message4 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : function (value, max) {
        return 'Please enter a value less than or equal to ' + max + '.';
    };

    return custom(function (value) {
        return !isNaN(value) && value <= max;
    }, { id: 'max', message: function message(value) {
            return _message4(value, max);
        } });
}

function range(min, max) {
    var _message5 = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : function (value, min, max) {
        return 'Please enter a value between ' + min + ' and ' + max + ' inclusive.';
    };

    return custom(function (value) {
        return !isNaN(value) && value >= min && value <= max;
    }, { id: 'range', message: function message(value) {
            return _message5(value, min, max);
        } });
}

export function digits() {
    var message = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : function (value) {
        return 'Please enter only digits.';
    };

    return custom(function (value) {
        return (/^\d*$/.test(value)
        );
    }, message());
}

/**
 * @param {bool} negative
 * @param message
 * @returns {{id, isAsync, message, validate, precondition, dependsOn, isOptional, type, compare}|*}
 */
export function realNumber() {
    var negative = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;
    var message = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : function (value) {
        return 'Please enter a valid number.';
    };

    return custom(function (value) {
        return (!negative || value >= 0) && /^-?(?:\d+|\d{1,3}(?:,\d{3})+)?(?:\.\d+)?$/.test(value);
    }, message());
}

// TODO: add rest from https://jqueryvalidation.org/documentation/#link-list-of-built-in-validation-methods